package GT::Maildir::FolderTree;

use base 'GT::Base';
use vars qw/$ERRORS $error *pwd *cd/;
use strict;
use warnings;

use GT::Maildir::Folders;
use GT::Maildir::FolderTree::Dir;

use Carp qw/croak/;

$error = '';
$ERRORS = {
    RENAME => "Could not rename '%s' to '%s'; Reason: '%s'",
    NOFOLDER => "The folder specified does not exist: %s",
    NOMESSAGE => "The message specified does not exist: %s",
    DELETE_ROOT => "You can not delete the root folder",
    DIR_EXISTS => "Folder %s already exists"
};

sub init {
    my $self = shift;
    my $class = ref $self;
    my $maildir = shift;
    croak "No maildir object specified to $class->new" unless $maildir;
    croak "Unknown options specified to $class->new: ", join(", ", @_) if @_;

    $self->{maildir} = $maildir;
    $self->{path} = $maildir->get_maildir_path;
    $self->{folders} = new GT::Maildir::Folders($maildir);
    $self->{cwd} = '.';
}

sub folders {
    return $_[0]->{folders};
}

sub convert_path {
    my $self = shift;
    my $dir = shift;
    croak "Invalid directory specified to $self->convert_path" unless defined $dir and length $dir;
    croak "Unknown arguments specified to $self->convert_path", join(", ", @_) if @_;
    if ($dir =~ /^=([.][\w. -]*)/) {
        return "$1";
    }
    my @elements = ($dir =~ m{/} ? split(m{/}, $dir) : $dir);
    my @path;
    if ($dir =~ m{^/}) {
        @path = ('');
    }
    else {
        @path = split /[.]/, $self->{cwd};
    }
    for (@elements) {
        next unless length $_;
        next if $_ eq '.';
        if ($_ eq '..') {
            pop @path if @path > 1;
        }
        else {
            push @path, $_;
        }
    }
    if ((@path == 1 and ($path[0] eq '' or $path[0] eq '.')) or !@path) {
        return ".";
    }
    elsif (@path == 1) {
        return "." . $path[0];
    }
    else {
        return join ".", @path;
    }
}

sub rm {
    my $self = shift;
    my $message = shift;
    croak "Invalid directory specified to $self->rm" unless defined $message and length $message;
    my $folder = $self->{cwd};
    if ($message =~ m{(.+)/((?:cur|new)/.+)}) {
        $message = $2;
        $folder = $self->convert_path($1);
    }
    my $path = "$self->{path}/$folder/$message";
    unless (-e $path) {
        return $self->error("NOMESSAGE", "WARN", $path);
    }
    unlink $path
        or return $self->error("UNLINK", "WARN", $path, "$!");
    return 1;
}

sub mv {
    my $self = shift;
    my $message = shift;
    croak "Invalid directory specified to $self->mv" unless defined $message and length $message;
    my $folder = $self->{cwd};
    if ($message =~ m{(.+)/((?:cur|new)/.+)}) {
        $message = $2;
        $folder = $1;
    }
    my $from_path = "$self->{path}/$folder/$message";
    my $to_folder = $self->convert_path(shift);
    my $to_path = "$self->{path}/$to_folder";

    return 2 if $folder eq $to_folder;

    unless (-e $from_path) {
        return $self->error("NOMESSAGE", "WARN", $from_path);
    }
    unless (-d $to_path) {
        return $self->error("NOFOLDER", "WARN", $to_path);
    }
    $to_path .= "/$message";
    unless (rename($from_path, $to_path)) {
        return $self->error("RENAME", "WARN", $from_path, $to_path, "$!");
    }
    return 1;
}

sub mark_read {
    my $self = shift;
    return $self->mark('cur', @_);
}

sub mark_unread {
    my $self = shift;
    return $self->mark('new', @_);
}

sub mark {
    my ($self, $st, $message) = @_;
    croak "Invalid directory specified to $self->mark"
        unless defined $message and length $message;

    my $anti = $st eq 'cur' ? 'new' : 'cur';
    my $folder = $self->{cwd};
    if ($message =~ m{(.+)/((?:cur|new)/.+)}) {
        $message = $2;
        $folder = $1;
        if ($message =~ m{^$st/}) {
            return 1;
        }
    }
    my $from_path = "$self->{path}/$folder/$message";
    unless (-e $from_path) {
        return $self->error("NOMESSAGE", "WARN", $from_path);
    }
    $message =~ s/^$anti/$st/;
    my $to_path = "$self->{path}/$folder/$message";
    unless (rename($from_path, $to_path)) {
        return $self->error("RENAME", "WARN", $from_path, $to_path, "$!");
    }
    return 1;
}

sub flag_answered {
    my $self = shift;
    return $self->flag('R', @_);
}

sub flag_seen {
    my $self = shift;
    return $self->flag('S', @_);
}

sub flag_draft {
    my $self = shift;
    return $self->flag('D', @_);
}

sub flag_deleted {
    my $self = shift;
    return $self->flag('T', @_);
}

sub flag {
    my ($self, $flag, $message) = @_;
    croak "Invalid directory specified to $self->flag"
        unless defined $message and length $message;

    my $folder = $self->{cwd};
    my $flags;
    if ($message =~ m{(?:([^/]+)/)?((?:new|cur)/[^/]+,S=\d+(?::\d+,(\w*))?)$}) {
        $folder = $1 if $1;
        $flags = $3 || '';
        if ($flags =~ m{$flag}i) {
            return $message;
        }
        $message = $2;
    }
    else {
        croak "Invalid message '$message'";
    }
    my $from_path = "$self->{path}/$folder/$message";
    unless (-e $from_path) {
        return $self->error("NOMESSAGE", "WARN", $from_path);
    }
    if ($message =~ /:\d+,\w*$/) {
        $message .= uc $flag;
    }
    else {
        $message .= ":2,$flag";
    }
    my $to_path = "$self->{path}/$folder/$message";
    unless (rename($from_path, $to_path)) {
        return $self->error("RENAME", "WARN", $from_path, $to_path, "$!");
    }
    return $message;
}

sub rmdir {
    my $self = shift;
    my $dir = shift;
    croak "Invalid directory specified to $self->rmdir" unless defined $dir and length $dir;
    my $force = shift;
    croak "Unknown arguments specified to $self->rmdir: ", join(", ", @_) if @_;
    my $orig_dir = $dir;
    $dir = $self->convert_path($dir);
    unless ($self->{folders}->get_folder($dir)) {
        return $self->error("NOFOLDER", "WARN", "$dir");
    }
    if ($dir eq '/') {
        return $self->error("DELETE_ROOT", "WARN");
    }
    if ($self->{cwd} eq $dir) {
        $self->chdir("..");
    }
    # Collect a list of folders that need updating. i.g. if we delete
    # .foo.bar this will return .foo.bar.baz, .foo.bar.bat etc
    my @del;
    {
        my (@dirs, @objs);
        push @objs, new GT::Maildir::FolderTree::Dir(
            Debug       => $self->{_debug},
            Directory   => $dir,
            Folders     => $self->{folders},
            ListFolders => 1
        );
        push @dirs, $dir;
        my ($d, $obj);
        while (($d = shift(@dirs) and $obj = shift(@objs))) {
            if (-d "$self->{path}/$d") {
                push @del, $d;
            }
            while (my $f = $obj->readdir) {
                push @objs, new GT::Maildir::FolderTree::Dir(
                    Debug       => $self->{_debug},
                    Directory   => "$d.$f",
                    Folders     => $self->{folders},
                    ListFolders => 1
                );
                push @dirs, "$d.$f";
            }
        }
    }
    my $errors = 0;
    unless ($force) {
        my $dir = do { local *FH; \*FH };
        for (@del) {
            for my $d (qw/cur new tmp/) {
                opendir $dir, "$self->{path}/$_/$d"
                    or return $self->error("OPENDIR", "WARN", "$self->{path}/$_/$d", "$!");
                while (my $f = readdir $dir) {
                    next if $f eq "." or $f eq "..";
                    return $self->error("RMDIR", "WARN", "$self->{path}/$_/$d", "directory not empty");
                }
                closedir $dir;
            }
        }
    }
    for (@del) {
        if ($_ eq $self->{cwd}) {
            $self->chdir("/");
        }
        my $dir = do { local *FH; \*FH };
        my $errors_here;
        for my $d (qw/cur new tmp/) {
            if (opendir $dir, "$self->{path}/$_/$d") {
                while (my $f = readdir $dir) {
                    next if $f eq "." or $f eq '..';
                    unlink "$self->{path}/$_/$d/$f" or do {
                        warn "Could not unlink $self->{path}/$_/$d/$f";
                    }
                }
                closedir $dir;
            }
            rmdir "$self->{path}/$_/$d" or do {
                $errors++;
                $errors_here = 1;
                $self->error("RMDIR", "WARN", "$self->{path}/$_/$d", "$!");
            }
        }
        next if $errors_here;
        if (opendir $dir, "$self->{path}/$_") {
            # Only go one level deep for courier imap files
            while (my $f = readdir $dir) {
                next if $f eq "." or $f eq '..';
                if (-f "$self->{path}/$_/$f") {
                    unlink "$self->{path}/$_/$f" or do {
                        warn "Could not unlink $self->{path}/$_/$f: $!";
                    };
                }
                elsif (-d _) {
                    my $sdir = do { local *SFH; \*SFH };
                    if (opendir $sdir, "$self->{path}/$_/$f") {
                        while (my $f = readdir $dir) {
                            next if $f eq "." or $f eq '..';
                            unlink "$self->{path}/$_/$f" or do {
                                warn "Could not unlink $self->{path}/$_/$f: $!";
                            };
                        }
                        closedir $sdir;
                    }
                    rmdir "$self->{path}/$_/$f" or do {
                        warn "Could not rmdir $self->{path}/$_/$f";
                    };
                }
            }
            closedir $dir;
        }
        rmdir "$self->{path}/$_" or do {
            $self->error("RMDIR", "WARN", "$self->{path}/$_", "$!");
            $errors++;
        };
        $self->{folders}->check_update($_);
    }
    return $errors ? undef : 1;
}

sub is_folder {
    my $self = shift;
    my $dir = shift;
    croak "Invalid directory specified to $self->is_folder" unless defined $dir and length $dir;
    croak "Unknown arguments specified to $self->is_folder: ", join(", ", @_) if @_;
    $dir = $self->convert_path($dir);
    my $folder = $self->{folders}->get_folder($dir);
    if (!$folder) {
        return 0;
    }
    if ($folder->is_deleted) {
        return 0;
    }
    return 1;
}

sub folder {
    my $self = shift;
    my $dir = shift;
    croak "Invalid directory specified to $self->folder" unless defined $dir and length $dir;
    croak "Unknown arguments specified to $self->folder: ", join(", ", @_) if @_;
    $dir = $self->convert_path($dir);
    my $folder = $self->{folders}->get_folder($dir);
    return $folder;
}

sub chdir {
    my $self = shift;
    my $dir = shift;
    croak "Invalid directory specified to $self->chdir" unless defined $dir and length $dir;
    croak "Unknown arguments specified to $self->chdir: ", join(", ", @_) if @_;
    $dir = $self->convert_path($dir);
    return _chdir($self, $dir);
}
*cd = \&GT::Maildir::FolderTree::chdir;

sub _chdir {
    my ($self, $dir) = @_;
    my $found = 0;
    if ($self->{folders}->get_folder($dir)) {
        $found = 1;
    }
    else {
        my $folders = $self->{folders}->get_list || [];
        my $match = $dir;
        for (@$folders) {
            if (/^\Q$match\E/) {
                $found = 1;
                last;
            }
        }
    }
    if (!$found) {
        return $self->error("NOFOLDER", "WARN", $dir);
    }
    $self->{cwd} = $dir;
    return 1;
}

sub mkdir {
    my $self = shift;
    my $dir = shift;
    croak "Invalid directory specified to $self->mkdir" unless defined $dir and length $dir;
    croak "Unknown arguments specified to $self->mkdir: ", join(", ", @_) if @_;
    my $folder = $self->convert_path($dir);
    if (-d "$self->{path}/$folder") {
        return $self->error("DIR_EXISTS", "WARN", $folder);
    }
    mkdir "$self->{path}/$folder", 0700
        or return $self->error("MKDIR", "WARN", "$self->{path}/$folder", "$!");
    for (qw/tmp new cur/) {
        mkdir "$self->{path}/$folder/$_", 0700
            or return $self->error("MKDIR", "WARN", "$self->{path}/$folder", "$!");
    }
    $self->{folders}->add_existing($folder);
    return 1;
}

sub cwd {
    my $self = shift;
    croak "Unknown arguments specified to $self->cwd: ", join(", ", @_) if @_;
    my @path = split /[.]/, $self->{cwd};
    if (@path <= 1) {
        return "/";
    }
    else {
        return join "/", @path;
    }
}
*pwd = \&GT::Maildir::FolderTree::cwd;

sub cwf {
    my $self = shift;
    croak "Unknown arguments specified to $self->cwd: ", join(", ", @_) if @_;
    my @path = split /[.]/, $self->{cwd};
    if (@path <= 1) {
        return "/";
    }
    else {
        return $path[$#path];
    }
}

sub opendir {
    my $self = shift;
    my ($dir, $cur, $new, $folders, $search_index, $search_sort, $offset, $limit);
    if (@_ == 1) {
        $dir = shift;
        $dir = "." unless defined $dir;
        croak "Invalid directory specified to $self->opendir" unless length $dir;
        croak "Unknown arguments to $self->opendir: ", join(", ", @_) if @_;
        $folders = $cur = $new = 1;
    }
    elsif (!@_ or !(@_ & 1)) {
        my %opts = @_;
        $opts{ lc $_ } = delete $opts{$_} for keys %opts;
        $dir = delete $opts{directory};
        $dir = '.' unless defined $dir;
        $cur = delete $opts{listcurmessages};
        $new = delete $opts{listnewmessages};
        $folders = delete $opts{listfolders};
        $search_index = delete $opts{searchindex};
        $search_sort = delete $opts{searchsort};
        $offset = delete $opts{offset};
        $limit = delete $opts{limit};
        croak "Unknown arguments to $self->opendir: ", join(", ", map { $_ . " => " . $opts{$_} } keys %opts) if keys %opts;
    }
    else {
        croak "Invalid arguments to $self->opendir: ", join(", ", @_);
    }
    return new GT::Maildir::FolderTree::Dir(
        Debug               => $self->{_debug},
        Directory           => $self->convert_path($dir),
        Folders             => $self->{folders},
        Limit               => $limit,
        ListCurrentMessages => $cur,
        ListFolders         => $folders,
        ListNewMessages     => $new,
        Offset              => $offset,
        SearchIndex         => $search_index,
        SearchSort          => $search_sort,
    );
}

sub deliver {
    my $self = shift;
    my $mail = shift;
    my $quotastr = shift;
    local $GT::Maildir::ERROR_MESSAGE = "GT::Maildir::FolderTree";
    return $self->{maildir}->deliver_message($self->{cwd}, $mail, $quotastr, @_);
}

sub _cwd {
    my $self = shift;
    croak "Unknown arguments specified to $self->_cwd: ", join(", ", @_) if @_;
    return $self->{cwd};
}

1;

