package GT::Maildir::Folders;

use base 'GT::Base';
use vars qw/$ERRORS $error/;
use strict;
use warnings;

use GT::Maildir::Folder;

use Carp qw/croak/;

$error = '';
$ERRORS = {
    RENAME => "Could not rename '%s' to '%s': %s",
    BAD_IDX => "Invalid folder index file '%s': '%s'",
    NOFOLDER => "You are trying to add a folder that does not exist on disk: %s",
};

sub init {
    my $self = shift;
    my $class = ref $self;
    my $maildir = shift;
    croak "No maildir object specified to $class->new" unless $maildir;

    croak "Unknown options specified to $class->new: ", join(", ", @_) if @_;

    my $path = $maildir->get_maildir_path;
    $self->{path} = $path;
    $self->{idx_path} = $maildir->get_config_path
        . "/folders.txt";
    $self->{folders} = {};
    $self->{locker} = ref($maildir->get_locker)->new;
    $self->{read_cache} = 1;
}

sub get_list {
    my $self = shift;
    croak "Unknown arguments to $self->get_list: ", @_ if @_;

    if ($self->{read_cache}) {
        $self->read_cache or return undef;
    }
    return [
        sort grep {
            !$self->{folders}{$_}->get('deleted')
        }
        map {
            $self->check_update($_);
            $_;
        } keys %{$self->{folders}}
    ];
}

sub get_new_count {
    my $self = shift;
    croak "Unknown arguments to $self->unread_count: ", @_ if @_;
    my $list = $self->get_list or return undef;
    my $total = 0;
    for (@$list) {
        my $folder = $self->get_folder($_) or next;
        $total += $folder->get_new_count;
    }
    return $total;
}

sub get_cur_count {
    my $self = shift;
    croak "Unknown arguments to $self->unread_count: ", @_ if @_;
    my $list = $self->get_list or return undef;
    my $total = 0;
    for (@$list) {
        my $folder = $self->get_folder($_) or next;
        $total += $folder->get_cur_count;
    }
    return $total;
}

sub get_new_list {
    my $self = shift;
    croak "Unknown options specified to $self->get_new_list: ", join(", ", @_) if @_;
    if ($self->{read_cache}) {
        $self->read_cache or return undef;
    }
    my $folders = _read_folders($self) or return undef;
    my @new_list;
    for (@$folders) {
        unless (exists $self->{folders}{$_}) {
            push @new_list, $_;
        }
    }
    return \@new_list;
}

sub add_existing {
    my $self = shift;
    my $folder = shift;
    croak "Invalid folder" unless defined $folder and length $folder;
    croak "Unknown arguments to $self->add_existing: ", join(", ", @_) if @_;
    unless (-d "$self->{path}/$folder") {
        return $self->error("NOFOLDER", "WARN", "$self->{path}/$folder");
    }
    if ($self->{read_cache}) {
        $self->read_cache or return undef;
    }
    $self->{folders}{$folder} = new GT::Maildir::Folder($self->{path}, $folder);
    $self->{folders}{$folder}->update;
    $self->write_cache;
}

sub get_folder {
    my $self = shift;
    croak "No folder specified to $self->folder_info" unless @_;
    my $folder = shift;
    croak "Invalid folder specified to $self->folder_info" unless defined $folder and length $folder;
    croak "Unknown options specified to $self->folder_info: ", join(", ", @_) if @_;
    if ($self->{read_cache}) {
        $self->read_cache or return undef;
    }
    return $self->{folders}{$folder};
}

sub needs_update {
    my $self = shift;
    croak "No folder specified to $self->needs_update" unless @_;
    my $folder_name = shift;
    croak "Invalid folder specified to $self->needs_update" unless defined $folder_name;
    my $folder = ref $folder_name ? $folder_name : $self->{folders}{$folder_name};
    croak "Unknown options specified to $self->needs_update: ", join(", ", @_) if @_;
    return undef unless $folder;
    $self->check_update($folder);

    if ($folder->get('needs_update')) {
        return 1;
    }
    elsif ($folder->get('deleted')) {
        return 2;
    }
    return 0;
}

sub check_update {
    my $self = shift;
    croak "No folder specified to $self->check_update" unless @_;
    my $folder_name = shift;
    croak "Invalid folder specified to $self->check_update" unless defined $folder_name;
    my $folder = ref $folder_name ? $folder_name : $self->{folders}{$folder_name};
    croak "Unknown options specified to $self->check_update: ", join(", ", @_) if @_;
    return undef unless $folder;

    if ($self->{read_cache}) {
        $self->read_cache or return undef;
    }
    my $needs_update = $folder->needs_update;
    if ($needs_update == 1) {
        $folder->update;
        $folder->set(needs_update => 1);
        $self->write_cache;
    }
    elsif ($needs_update == 2) {
        $folder->set(deleted => 1);
        $self->write_cache;
    }
    return $needs_update;
}

sub clear_needs_update {
    my $self = shift;
    croak "No folder specified to $self->clear_needs_update" unless @_;
    my $folder_name = shift;
    croak "Invalid folder specified to $self->clear_needs_update" unless defined $folder_name;
    my $folder = ref $folder_name ? $folder_name : $self->{folders}{$folder_name};
    croak "Unknown options specified to $self->clear_needs_update: ", join(", ", @_) if @_;
    return undef unless $folder;

    if ($self->{read_cache}) {
        $self->read_cache or return undef;
    }
    if ($folder->get('needs_update')) {
        $folder->set(needs_update => 0);
    }
    elsif ($folder->get('deleted')) {
        delete $self->{folders}{$folder->get_name};
    }
    $self->write_cache;
    return 1;
}

sub read_cache {
    my $self = shift;
    croak "Unknown options specified to $self->read_cache: ", join(", ", @_) if @_;

    if (!-e $self->{idx_path}) {
        $self->update_cache or return undef;
    }
    $self->{locker}->sh_lock($self->{idx_path}, 60*10, 60*20);
    my $data = do "$self->{idx_path}";
    $self->{locker}->unlock($self->{idx_path});
    if (!$data) {
        return $self->error("BAD_IDX", "WARN", $self->{idx_path}, "$@") if $@;
        return $self->error("BAD_IDX", "WARN", $self->{idx_path}, "$!") unless defined $data;
        return $self->error("BAD_IDX", "WARN", $self->{idx_path}, "Bad format");
    }
    $self->{folders} = {};
    for (keys %$data) {
        my $folder = new GT::Maildir::Folder($self->{path}, $_);
        while (my ($k, $v) = each(%{$data->{$_}})) {
            $folder->set($k => $v);
        }
        $self->{folders}{$_} = $folder;
    }
    $self->{read_cache} = 0;
    return 1;
}

sub write_cache {
    my $self = shift;
    croak "Unknown options specified to $self->write_cache: ", join(", ", @_) if @_;

    my $new = !-e $self->{idx_path};
    my $data = {};
    for (keys %{$self->{folders}}) {
        $self->{folders}{$_}->set(needs_update => 1) if $new;
        $data->{$_} = $self->{folders}{$_}->data;
    }

    $self->{locker}->ex_lock("$self->{idx_path}.new", 60*10, 60*20);
    open my $fh, ">$self->{idx_path}.new" or do {
        $self->{locker}->unlock("$self->{idx_path}.new");
        return $self->error("WRITEOPEN", "WARN", $self->{idx_path}, "$!");
    };
    print $fh GT::Dumper->dump(
        structure => 1,
        data      => $data
    );
    close $fh;
    $self->{locker}->ex_lock($self->{idx_path}, 60*10, 60*20);
    rename "$self->{idx_path}.new", $self->{idx_path} or do {
        $self->{locker}->unlock($self->{idx_path});
        $self->{locker}->unlock("$self->{idx_path}.new");
        return $self->error("RENAME", "WARN", "$self->{idx_path}.new", $self->{idx_path}, "$!");
    };
    $self->{locker}->unlock($self->{idx_path});
    $self->{locker}->unlock("$self->{idx_path}.new");
    return 1;
}

# XXX this wipes any custom data for folders XXX
sub update_cache {
    my $self = shift;
    croak "Unknown options specified to $self->update_cahe: ", join(", ", @_) if @_;

    my $dir = do { local *FH; \*FH };
    my $folders =  _read_folders($self) or return undef;
    $self->{folders} = {};
    for (@$folders) {
        $self->{folders}{$_} = new GT::Maildir::Folder($self->{path}, $_);
        $self->{folders}{$_}->update;
    }
    $self->write_cache or return undef;
    return 1;
}

sub _read_folders {
    my ($self) = @_;
    my $dir = do { local *FH; \*FH };
    opendir $dir, $self->{path} or do {
        return $self->error("OPENDIR", "WARN", $self->{path}, "$!");
    };
    my @folders = grep { /^[.](?![.])/ and -d "$self->{path}/$_/cur" } readdir $dir;
    closedir $dir;
    return \@folders;
}

1;

