package GT::Maildir::Indexer;

use strict;
use warnings;
use base 'GT::Base';

use GT::Maildir qw(:status);

use Carp qw/croak/;

sub init {
	my $self = shift;
	my $class = ref $self;
	my $maildir = shift;
	croak "No maildir object specified to $class->new" unless $maildir;
	croak "Unknown options specified to $class->new: ", join(", ", @_) if @_;
	$self->{maildir} = $maildir;
}

sub get_maildir {
	my $self = shift;
	croak "Unknown arguments to $self->get_maildir: ", join(", ", @_) if @_;
	return $self->{maildir};
}

sub parse_message {
    my ($self, $folder, $status, $file, $flags) = @_;

    my $fl_answered = $flags =~ /R/ ? 1 : 0;
    my $fl_seen = $flags =~ /S/ ? 1 : 0;
    my $fl_draft = $flags =~ /D/ ? 1 : 0;
    my $fl_deleted = $flags =~ /T/ ? 1 : 0;

    my $path = "$self->{path}/$folder/$status/$file";

    # Parse the email first
    my $parser = new GT::Mail::Parse(
        in_file => $path
    );
    my $head = $parser->parse;
    my $in_date = $head->get('Date');
    $in_date =~ tr/\r\n//d if $in_date;
    my $date = $self->parse_date($in_date);
    if (!$date) {
        $date = time;
    }

    my $has_html = $head->mime_type =~ /html/ ? 1 : 0;
    my @parts = ($head);
    my $size = 0;

    # Keep track of whether or not we're still looking at the body section of an
    # email.  Hitting things like an attachment or a message/rfc822 content type
    # should mean that we're done with the body section.  This way, we can
    # correctly determine if the email has a html part or not.  The other thing we
    # look for is a multipart/alternative - once we find the first one, we assume
    # that it and its contents are in the body (either the body itself, or an
    # inline attachment).
    my ($in_body_part, $last_body_part) = (1, 0);
    for (my $i = 0; $i < @parts; $i++) {
        my @p = $parts[$i]->parts();
        splice @parts, $i + 1, 0, @p if @p;
        $size += $parts[$i]->size() || 0;

        $in_body_part = 0 if $i > $last_body_part;
        next unless $in_body_part;
        $last_body_part += @p;
        my $mime_type = $parts[$i]->mime_type() || '';
        my $content_disp = lc $parts[$i]->mime_attr('content-disposition') || '';
        if ($content_disp =~ /attachment/ or $mime_type =~ /message\/rfc822/i) {
            $in_body_part = 0;
            $last_body_part = $i - 1 if $i > 0;
        }
        elsif ($mime_type =~ m{multipart/(?:alternative|related)}) {
            $last_body_part = $i + @p unless $last_body_part > $i + @p;
        }
        elsif ($mime_type =~ m{multipart/mixed}) {
            if (@p and $p[0]->mime_type() =~ m{text/(?:plain|html)|multipart/(?:alternative|related)}) {
                # When we encounter a multipart/mixed with a text or multipart first child, assume that
                # that child is the body.
                $last_body_part = $i + 1;
            }
            elsif (@p) {
                # We found a multipart/mixed with some unknown type as the first part that we can't
                # handle; make the body end just before this part.
                $last_body_part = $i - 1;
            }
        }
        elsif ($mime_type =~ /html/) {
            $has_html = 1;
        }
    }

    # Get the information ready for insertion
    my @reply = $head->get('Reply-To');
    my @from = $head->get('From');
    @from = @reply unless @from;

    my $has_attachment = ($head->get('Content-Type') and $head->get('Content-Type') =~ m{multipart/mixed}i) ? 1 : 0;
    my %to;
    for my $field (qw(to from cc bcc reply_to)) {
        my @addresses = map {
            GT::Mail::Parts->split_line('\s*,\s*', $_)
        } ($field eq 'from'
            ? @from
            : $field eq 'reply_to'
                ? @reply
                : $head->get($field));
        $to{$field} = [];
        $to{"${field}_name"} = [];
        $to{"${field}_email"} = [];
        for my $addr (@addresses) {
            my ($name, $email) = GT::Mail->parse_address($addr);
            push @{ $to{$field} }, $addr;
            push @{ $to{"${field}_name"} }, $name;
            push @{ $to{"${field}_email"} }, $email;
        }
    }
    my %email;
    for my $field (qw(to from cc bcc reply_to)) {
        $email{$field} = join(", ", @{ $to{$field} }) || '';
        $email{"${field}_sort"} = join(", ", @{ $to{$field} }) || '';
        $email{"${field}_name"} = join(", ", @{ $to{"${field}_name"} }) || '';
        $email{"${field}_name_sort"} = join(", ", @{ $to{"${field}_name"} }) || '';
        $email{"${field}_email"} = join(", ", @{ $to{"${field}_email"} }) || '';
        $email{"${field}_email_sort"} = join(", ", @{ $to{"${field}_email"} }) || '';
    }
    $email{flag_answered} = $fl_answered;
    $email{flag_seen} = $fl_seen;
    $email{flag_draft} = $fl_draft;
    $email{flag_deleted} = $fl_deleted;
    $email{folder} = $folder;
    $email{status} = ($status eq "new" ? ST_NEW : ST_CUR);
    $email{filename} = $file;
    $email{path} = "$folder/$status/$file";
    $email{size} = $size;
    $email{date} = $date;
    $email{date_recieved} = time();
    $email{subject} = join(", ", $head->get('Subject')) || '';
    $email{subject_sort} = join(", ", $head->get('Subject')) || '';
    $email{message_id} = ($head->get('Message-Id') || '<>');
    $email{has_attachment} = $has_attachment;
    $email{has_html} = $has_html;
    $email{content_type} = $head->get('Content-Type') || '';
    $email{dangerous} = (($head->get('X-GMail-Attachments') || '') eq 'filtered' ? 1 : 0);
    $email{dangerous} ||= (($head->get('X-AntiVirus') || '') =~ /Virus/i ? 1 : 0);

    # Spam assassin headers
    my $spam_status = $head->get('X-Spam-Status');
    my ($is_spam, $spam_score, $spam_score_required, $spam_learn) = (0, 0, 0, 0);
    if ($spam_status) {
        my $fields;
        ($is_spam, $fields) = split /,\s*/, $spam_status, 2;
        $is_spam = $is_spam =~ /Yes/i ? 1 : 0;
        my %fields = map { split '=', $_, 2 } split ' ', $fields;

        $spam_score = exists $fields{score} ? $fields{score} : 0;
        $spam_score_required = exists $fields{required} ? $fields{required} : 0;
        $spam_learn = exists $fields{autolearn}
            ? $fields{autolearn} =~ /ham/i ? 1 : 2
            : 0;
    }

    $email{spam_status} = $is_spam;
    $email{spam_score} = $spam_score;
    $email{spam_score_required} = $spam_score_required;
    $email{spam_learn} = $spam_learn;
    $email{spam_report} = $head->get('X-Spam-Report') || '';
    $email{spam_version} = $head->get('X-Spam-Checker-Version') || '';

    my ($header, $body) = ('', '');

    # Find the header for searching and the body for searching
    my $match;
    my $attach_index;
    for (my $i = 0; $i < @parts; $i++) {
        my $part = $parts[$i];
        $header .= " " . $part->header_as_string;
        my $type = $part->mime_type() || 'text/plain';
        next if $type =~ /multipart/i;


        my $disposition = lc $part->mime_attr('content-disposition') || '';
        if (($type =~ /plain/i or $type =~ /html/i) and $disposition ne 'attachment' and not $match) {
            my $msg_in = $part->body_in() or next;
            my $msg_data;
            if ($msg_in eq 'FILE') {
                $GT::Mail::Parts::error ||= '';
                my $fh = $part->open('r') or $self->error($GT::Mail::Parts::error);
                binmode $fh;
                read($fh, $msg_data, -s $fh);
                close $fh;
            }
            else {
                $msg_data = $part->body_data();
            }
            $msg_data ||= '';

            $match = 1;

            if ($type eq 'text/html') {
                $self->strip_html($msg_data);
            }
            $body = $msg_data;
            undef $msg_data;
        }
    }
    $email{body} = $body;
    $email{body_preview} = substr($body, 0, 500);
    $email{header} = $header;
    undef $body;
    undef $header;
    return \%email;
}

sub stip_html {
    my ($self) = @_;
    local *_ = $_[1];
    my $dash = ('-' x 60);
    s/\r?\n/ /g;
    s/\r/ /g;
    s/ +/ /g;

    s/&nbsp;/ /ig;
    s/&quot;/"/ig;
    s/&amp;/&/ig;
    s/&copy;/(c)/ig;
    s/&reg;/(r)/ig;
    s/&trade;/^TM/ig;

    s/<!--.*?-->//sg;
    s[<script[^>]*>.*?</script>][]sig;
    s[<style[^>]*>.*?</style>][]sig;
    s[<head[^>]*>.*?</head>][]sig;
    s[<li>\s*(.*?)\s*</li>][\n* $1\n]sig;
    s/<li>\s*([^<]*)/\n* $1/sig;
    s[</p[^>]*>][\n\n]ig;
    s/<br[^>]*>/\n/ig;
    s[</div[^>]*>][\n]ig;
    s[<b>([^<]*)</b>][*$1*]ig;
    s[<i>([^<]*)</i>][_$1_]ig;
    s[</blockquote[^>]*>][\n\n]ig;
    s/<hr[^>]*>/\n$dash\n/ig;

    my @tokens = split /(<[^>"']*(?:(?:(?:"[^"]*"[^>"]*)|(?:'[^']*'[^>']*)))*>)/, $_;
    $_ = join '' => map { $tokens[$_] } grep { not $_ % 2 } 0 .. $#tokens;

    s/&lt;/</ig;
    s/&gt;/>/ig;
    s/^ //mg;
}

sub parse_date {
    my ($self, $date) = @_;
    $date || return;

# Fix any dates where it doesn't specify the seconds
    $date =~ s/ (\d?\d:\d?\d)( |$)/ $1:00$2/;

    my $format;
    CASE: for ($date) {
        /\w+, \d\d? \w+ \d{4} \d?\d:\d?\d:\d\d (?:[+-]\d+|\w+)/    and $format = '%ddd%, %d% %mmm% %yyyy% %H%:%M%:%s% %o%',   last CASE;
        /\w+, \d\d? \w+ \d{4} \d?\d:\d?\d:\d\d "(?:[+-]\d+|\w+)"/  and $format = '%ddd%, %d% %mmm% %yyyy% %H%:%M%:%s% "%o%"', last CASE;
        /\w+, \d\d? \w+ \d{4} \d?\d:\d?\d:\d\d  (?:[+-]\d+|\w+)/   and $format = '%ddd%, %d% %mmm% %yyyy% %H%:%M%:%s%  %o%',  last CASE;
        /\w+, \d\d? \w+ \d{4} \d?\d:\d?\d:\d\d  "(?:[+-]\d+|\w+)"/ and $format = '%ddd%, %d% %mmm% %yyyy% %H%:%M%:%s%  "%o%"',last CASE;
        /\w+, \d\d? \w+ \d{4} \d?\d:\d?\d:\d\d/                    and $format = '%ddd%, %d% %mmm% %yyyy% %H%:%M%:%s%',       last CASE;
        /\w+, \d\d? \w+ \d{2} \d?\d:\d?\d:\d\d/                    and $format = '%ddd%, %d% %mmm% %yy% %H%:%M%:%s%',         last CASE;

        /\w+,  \d\d? \w+ \d{4} \d?\d:\d?\d:\d\d (?:[+-]\d+|\w+)/   and $format = '%ddd%,  %d% %mmm% %yyyy% %H%:%M%:%s% %o%',  last CASE;
        /\w+,  \d\d? \w+ \d{4} \d?\d:\d?\d:\d\d "(?:[+-]\d+|\w+)"/ and $format = '%ddd%,  %d% %mmm% %yyyy% %H%:%M%:%s% "%o%"',last CASE;
        /\w+,  \d\d? \w+ \d{4} \d?\d:\d?\d:\d\d/                   and $format = '%ddd%,  %d% %mmm% %yyyy% %H%:%M%:%s%',      last CASE;
        /\w+,  \d\d? \w+ \d{2} \d?\d:\d?\d:\d\d/                   and $format = '%ddd%,  %d% %mmm% %yy% %H%:%M%:%s%',        last CASE;

        /\d\d? \w+ \d{4} \d?\d:\d?\d:\d\d (?:[+-]\d+|\w+)/         and $format = '%d% %mmm% %yyyy% %H%:%M%:%s% %o%',          last CASE;
        /\d\d? \w+ \d{4} \d?\d:\d?\d:\d\d "(?:[+-]\d+|\w+)"/       and $format = '%d% %mmm% %yyyy% %H%:%M%:%s% "%o%"',        last CASE;
        /\d\d? \w+ \d{4} \d?\d:\d?\d:\d\d  (?:[+-]\d+|\w+)/        and $format = '%d% %mmm% %yyyy% %H%:%M%:%s%  %o%',         last CASE;
        /\d\d? \w+ \d{4} \d?\d:\d?\d:\d\d  "(?:[+-]\d+|\w+)"/      and $format = '%d% %mmm% %yyyy% %H%:%M%:%s%  "%o%"',       last CASE;
        /\d\d? \w+ \d{2} \d?\d:\d?\d:\d\d [AaPpMm]{2}/             and $format = '%d% %mmm% %yy% %h%:%M%:%s% %tt%',           last CASE;
    }
    $format or return;
    return GT::Date::timelocal(GT::Date::parse_format($date, $format));
}

1;


