package GT::Maildir::Lucene::Indexer;

use strict;
use warnings;
use base 'GT::Maildir::Indexer';
use vars qw/$ERRORS $ERROR_MESSAGE/;

sub LOCK_BLOCK () { 60 * 10 } # 10 minutes
sub LOCK_STALE () { 60 * 20 } # 20 minutes

use GT::Date;
use GT::Dumper;
use GT::File::Tools qw/filename/;
use GT::Mail::Parse;
use GT::Mail::Parts;
use GT::Mail;
use GT::Maildir qw/:status/;
use GT::Maildir::Lucene qw/%Lucene_Table_Defs/;
use GT::Maildir::Lucene::Indexer::Analyzer;

use Lucene;
use Carp qw/croak/;

$ERRORS = {
    CORRUPT_INDEX => "Corrupt Lucene index: %s"
};
$ERROR_MESSAGE = "GT::Maildir::Indexer";

sub init {
    my $self = shift;
    $self->SUPER::init(@_);
    my $path = $self->get_maildir->get_config_path . "/lucene";
    unless (-d $path) {
        unlink $path;
        mkdir $path, 0700 or $self->error("MKDIR", "FATAL", $path, "$!");
        $self->{create_idx} = 1;
    }
    unless (Lucene::Index::IndexReader->indexExists($path)) {
        $self->{create_idx} = 1;
    }
    $self->{db_path} = $path;
    $self->{path} = $self->get_maildir->get_maildir_path;
    $self->{locker} = ref($self->get_maildir->get_locker)->new;
}

sub is_new {
    my $self = shift;
    croak "Unknown arguments to $self->is_new: ", join(", ", @_) if @_;
    return $self->{create_idx};
}

sub delete_message {
    my $self = shift;
    my $path = shift;
    croak "Invalid path specified to $self->delete_message" unless defined $path and length $path;
    my ($folder, $status, $file);
    if ($path =~ m{([^/]+)/(new|cur)/([^/]+)$}) {
        $folder = $1;
        $status = $2;
        $file = $3;
    }
    else {
        croak "Invalid path specified to $self->delete_message";
    }
    croak "Unknown arguments to $self->delete_message: ", join(", ", @_) if @_;

    $self->{locker}->ex_lock($self->{db_path}, LOCK_BLOCK, LOCK_STALE);
    my $store = GT::Maildir::Lucene->store($self->{db_path});
    if (!$store) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$GT::Maildir::Lucene::error");
    }
    my $reader = eval {
        local $SIG{__DIE__};
        if ($self->{create_idx}) {
            $self->{create_idx} = 0;
            my $analyzer = new GT::Maildir::Lucene::Indexer::Analyzer;
            my $writer = new Lucene::Index::IndexWriter($store, $analyzer, 1);
            $writer->close();
        }
        Lucene::Index::IndexReader->open($store);
    };
    if ($@) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$@");
    }
    my $term = new Lucene::Index::Term(path => "$folder/$status/$file");
    $reader->deleteDocuments($term);
    $reader->close;
    undef $reader;
    $self->{locker}->unlock($self->{db_path});
    return 1;
}

sub delete_folder {
    my $self = shift;
    my $folder = shift;
    croak "Invalid folder specified to $self->delete_folder" unless defined $folder and length $folder;
    croak "Unknown arguments to $self->delete_folder: ", join(", ", @_) if @_;

    $self->{locker}->ex_lock($self->{db_path}, LOCK_BLOCK, LOCK_STALE);
    my $store = GT::Maildir::Lucene->store($self->{db_path});
    if (!$store) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$GT::Maildir::Lucene::error");
    }
    my $reader = eval {
        local $SIG{__DIE__};
        if ($self->{create_idx}) {
            $self->{create_idx} = 0;
            my $analyzer = new GT::Maildir::Lucene::Indexer::Analyzer;
            my $writer = new Lucene::Index::IndexWriter($store, $analyzer, 1);
            $writer->close();
        }
        Lucene::Index::IndexReader->open($store);
    };
    if ($@) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$@");
    }
    my $term = new Lucene::Index::Term(folder => $folder);
    $reader->deleteDocuments($term);
    $reader->close;
    undef $reader;
    $self->{locker}->unlock($self->{db_path});
    return 1;
}

sub update {
    my ($self, $path, $new) = @_;

    # Ensure we have a proper index
    my ($store, $analyzer, $searcher);
    $store = GT::Maildir::Lucene->store($self->{db_path});
    if (!$store) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$GT::Maildir::Lucene::error");
    }
    $self->{locker}->sh_lock($self->{db_path}, LOCK_BLOCK, LOCK_STALE);
    eval {
        local $SIG{__DIE__};
        $analyzer = new GT::Maildir::Lucene::Indexer::Analyzer;
        if ($self->{create_idx}) {
            my $writer = new Lucene::Index::IndexWriter($store, $analyzer, 1);
            $writer->close();
            $self->{create_idx} = 0;
        }
        $searcher = new Lucene::Search::IndexSearcher($store);
    };
    if ($@) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$@");
    }

    my $parser = new Lucene::QueryParser('path', $analyzer);
    my $lucene_query = eval { $parser->parse(qq{path: "$path"}) };
    if ($@) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("INVALID_QUERY", "WARN", "$@");
    }
    my $hits = $searcher->search($lucene_query);
    $self->{locker}->unlock($self->{db_path});

    $self->delete_message($path) or return undef;

    my $table_defs = GT::Maildir::Lucene->get_table_defs;
    for (my $i = 0; $i < $hits->length; ++$i) {
        my $doc = $hits->doc($i);
        my $document = new Lucene::Document;
        for (keys %$table_defs) {
            my $cp = $_;
            $cp =~ s/_sort$//;
            if (exists $new->{$cp}) {
                if ($_ eq 'data') {
                    $new->{$_} = GT::Dumper->dump_structure($new->{$_});
                }
                $document->add($self->field_for($_, $new->{$cp}));
                if ($_ ne $cp) {
                    $document->add($self->field_for($cp, $new->{$cp}));
                }
            }
            else {
                $document->add($self->field_for($_, $doc->get($_)));
            }
        }
        $self->{locker}->ex_lock($self->{db_path}, LOCK_BLOCK, LOCK_STALE);
        eval {
            local $SIG{__DIE__};
            my $writer = new Lucene::Index::IndexWriter($store, $analyzer, $self->{create_idx});
            $writer->addDocument($document);
            $writer->optimize();
            $writer->close();
        };
        if ($@) {
            $self->{locker}->unlock($self->{db_path});
            return $self->error("CORRUPT_INDEX", "WARN", "$@");
        }
        $self->{locker}->unlock($self->{db_path});
    }
}

sub index {
    my $self = shift;
    my $path = shift;
    my $data = shift || {};
    warn "Indexing '$path'";
    croak "Invalid path specified to $self->index" unless defined $path and length $path;
    my ($folder, $status, $file, $flags);
    if ($path =~ m{([^/]+)/(new|cur)/([^/]+,S=\d+(?::\d+,(\w*))?)$}) {
        $folder = $1;
        $status = $2;
        $file = $3;
        $flags = $4 || '';
    }
    else {
        croak "Invalid maildir file '$path'";
    }
    unless ($self->{create_idx}) {
        $self->delete_message("$folder/$status/$file") or return undef;
    }

    my $document = $self->email_to_document($folder, $status, $file, $flags);

    $document->add($self->field_for(data => GT::Dumper->dump_structure($data)));

    $self->{locker}->ex_lock($self->{db_path}, LOCK_BLOCK, LOCK_STALE);
    my $store = GT::Maildir::Lucene->store($self->{db_path});
    if (!$store) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$GT::Maildir::Lucene::error");
    }
    my $analyzer = eval {
        local $SIG{__DIE__};
        new GT::Maildir::Lucene::Indexer::Analyzer;
        
    };
    if ($@) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$@");
    }
    my $writer;
    eval {
        local $SIG{__DIE__};
        $writer = new Lucene::Index::IndexWriter($store, $analyzer, $self->{create_idx});
        $self->{create_idx} = 0;
        $writer->addDocument($document);
        $writer->optimize();
    };
    if ($@) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$@");
    }
    $writer->close();
    undef $writer;
    $self->{locker}->unlock($self->{db_path});
    return 1;
}

sub field_for {
    my ($self, $name, $value) = @_;
    my $method = $Lucene_Table_Defs{$name}{storage};
    croak "No method for $name!\n" unless $method;
    croak "Undefined name specified to $self->field_for" unless defined $name;
    croak "Undefined value specified to $self->field_for" unless defined $value;
    return Lucene::Document::Field->$method($name, $value);
}

sub email_to_document {
    my $self = shift;

    my $email = $self->parse_message(@_);

    # Get the information ready for insertion
    my $document = new Lucene::Document;
    while (my ($f, $v) = each %$email) {
        $document->add($self->field_for($f => $v));
    }
    return $document;
}


1;

