package GT::Maildir::Lucene::Searcher;

use strict;
use warnings;
use base 'GT::Maildir::Searcher';
use vars qw/$ERRORS $ERROR_MESSAGE/;

sub LOCK_BLOCK () { 60 * 10 } # 10 minutes
sub LOCK_STALE () { 60 * 20 } # 20 minutes

use GT::Maildir::Search::Sort;
use GT::Maildir::Search::Query;
use GT::Maildir::Lucene::Searcher::Results;
use GT::Maildir::Lucene::Indexer::Analyzer;
use GT::Maildir::Lucene qw/:sort_type/;

use Lucene;
use Carp qw/croak/;

$ERRORS = {
    CORRUPT_INDEX => "Corrupt Lucene index: %s",
    INVALID_QUERY => "Invalid Lucene query: %s"
};
$ERROR_MESSAGE = "GT::Maildir::Searcher";

sub init {
    my $self = shift;
    $self->SUPER::init(@_);
    my $path = $self->get_maildir->get_config_path 
        . "/lucene";
    unless (-d $path) {
        unlink $path;
        mkdir $path, 0700 or $self->error("MKDIR", "FATAL", $path, "$!");
        $self->{create} = 1;
    }
    unless (Lucene::Index::IndexReader->indexExists($path)) {
        $self->{create} = 1;
    }
    $self->{db_path} = $path;
}

sub search {
    my $self = shift;
    my $class = ref $self;

    croak "Invalid arguments to $self->search. Arguments must be key/value pairs" if @_ & 1;
    my %opts = @_;
    $opts{ lc $_ } = delete $opts{$_} for keys %opts;

    my $default_field = delete $opts{defaultfield};
    $default_field = GT::Maildir::Lucene->get_default_field unless $default_field;

    my $query = delete $opts{query};
    my $query_string = $query ? $class->query_string($query) : "";

    my $sort = delete $opts{sort};
    $sort = GT::Maildir::Search::Sort->RELEVANCE unless $sort;

    # Ensure we have a proper index
    my ($store, $analyzer, $searcher);
    $store = GT::Maildir::Lucene->store($self->{db_path});
    if (!$store) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$GT::Maildir::Lucene::error");
    }
    $self->{locker}->sh_lock($self->{db_path}, LOCK_BLOCK, LOCK_STALE);
    eval {
        local $SIG{__DIE__};
        $analyzer = new GT::Maildir::Lucene::Indexer::Analyzer;
        if ($self->{create}) {
            $self->{create} = 0;
            my $writer = new Lucene::Index::IndexWriter($store, $analyzer, 1);
            $writer->close();
        }
        $searcher = new Lucene::Search::IndexSearcher($store);
    };
    if ($@) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("CORRUPT_INDEX", "WARN", "$@");
    }

    # Figure out the sort order
    my @sortfields;
    {
        my @fields = ($sort);
        my $table_defs = GT::Maildir::Lucene->get_table_defs;
        while (@fields) {
            my $field = shift @fields;
            if ($field->is_group) {
                push @fields, @{ $field->get_sortfields };
            }
            elsif ($field->is_field) {
                my $name = $field->get_field;
                my $defs = $table_defs->{$name};
                if ($defs) {
                    my $field_name = exists($defs->{sort_field}) ? $defs->{sort_field} : $name;
                    my $sort_type = exists($defs->{sort_type}) ? $defs->{sort_type} : SB_AUTO;
                    push @sortfields, new Lucene::Search::SortField($field_name, $sort_type, $field->is_descending);
                }
            }
            elsif ($field->is_indexorder) {
                push @sortfields, Lucene::Search::SortField->FIELD_DOC;
            }
            elsif ($field->is_relevance) {
                push @sortfields, Lucene::Search::SortField->FIELD_SCORE;
            }
        }
    }
    my $lucene_sort = @sortfields ? new Lucene::Search::Sort(@sortfields) : Lucene::Search::Sort->RELEVANCE;

    # Parse the query
    my $parser = new Lucene::QueryParser($default_field, $analyzer);
    my $lucene_query = eval { $parser->parse($query_string) };
    if ($@) {
        $self->{locker}->unlock($self->{db_path});
        return $self->error("INVALID_QUERY", "WARN", "$@");
    }

    # Perform the search and return the results as a tied array
    my $hits = $searcher->search($lucene_query, $lucene_sort);
    my @results;
    tie @results, 'GT::Maildir::Lucene::Searcher::Results', $hits, $searcher;
    $self->{locker}->unlock($self->{db_path});
    return \@results;
}

sub query_string {
    my $class = shift;
    my $query = shift;
    croak "Unknown arguments to $class->query_string: ", @_ if @_;
    return "" unless $query;

    if ($query->can('get_terms')) {
        my $terms = $query->get_terms;
        my $bool_string = $query->get_bool == GQUERY_AND ? " AND " : " OR ";
        if (@$terms == 1) {
            return $class->query_string($terms->[0]);
        }
        return join($bool_string, map { '(' . $class->query_string($_) . ')' } @$terms);
    }
    else {
        my $field = $query->get_field;
        my $query_string = $query->get_query;
        return $field ? "$field:($query_string)" : "($query_string)";
    }
}

1;

